"""
Figure 9 of JFM 2019 vs PF2021
"""
import sys
sys.path.append("main")
from ODEdrop import *
from matplotlib.colors import LinearSegmentedColormap
import matplotlib
from similaritymeasures import frechet_dist

cases = ['fig10a','fig10b']

def frechet(C1,C2,closed=True):
    if closed:
        n1 = len(C1)
        n2 = len(C2)
        
        if n1>n2:
            n1,n2 = n2,n1
            C1,C2 = C2,C1
            
        C2 = np.vstack((C2,C2[0]))
        i = np.arange(n1)
        q = np.inf
        for j in range(n1):
            ids = np.hstack((np.roll(i,j),[-j]))
            q = np.min([q,frechet_dist(C1[ids],C2)])
            
        return q
    else:
        return frechet_dist(C1,C2)

def calculate_area(x, y):
    n = len(x)
    area = 0

    for i in range(n):
        j = (i + 1) % n
        area += (x[i] * y[j]) - (x[j] * y[i])

    return abs(area) / 2.0



def tikzplot(file,X,Y,style='',forget=True):
   with open(file,"w") as fid:
       if forget:
           style = style+',forget plot'

       if len(style)>0:
           fid.write('\\addplot['+style+']\n')
       else:
           fid.write('\\addplot\n')
       fid.write('table[row sep=crcr]{\n')
       dataXY = [X,Y]
       for x in zip(*dataXY):
           fid.write("%1.8f\t%1.8f\\\\\n" % (x))
       fid.write('};\n')


for case in cases:

    # DEFINE HETEROGENEITY
    if case == 'fig10a':
        def g(x,y):
            m1 = 20
            m2 = 1
            F = 0.5+0.3*np.tanh(m2*np.cos(np.pi*x))*np.tanh(m2*np.sin(np.pi*y))
            return 1.4+(1-0.5*(1-np.tanh(m1*(y+1))*np.tanh(m1*(y-1)))*(1-np.tanh(m1*(x-1)))/2)*np.arctan2(y,x)/np.pi*F+\
                       (1-np.tanh(m1*(y+1))*np.tanh(m1*(y-1)))*(1-np.tanh(m1*(x-1)))/10
    
    elif case == 'fig10b':
        def g(x,y):
            m1 = 1
            return 2 + 0.1*(x)- 0.2*(np.tanh(m1*np.cos(np.pi*(x+2*y))) + \
                                     np.tanh(m1*np.cos(np.pi*(2*x-y))))
      
    # DRAW SURFACE
    fig, ax = plt.subplots(figsize=(5,5))
    ax  = plt.axes()
    cmap = LinearSegmentedColormap.from_list('custom greys', [(1,1,1),(0.6,0.6,0.6)], N=256)
    plt.axis('equal')
    
    size_title  = 20
    size_labels = 18
    size_ticks  = 15
    size_legend = 16
    
    # Parameters & initialisation
    grid_n  = 128
    u       = np.arange(grid_n)*2*np.pi/grid_n
    
    if case == 'fig10a':
        t_end   = 2000
        Xc      = -3.0
        Yc      = +2.5
        n_arr   = 6
        slip    = 1e-3
        Tplot   = np.linspace(0,t_end,n_arr) 
    elif case == 'fig10b':
        t_end   = 300
        Xc      = +4
        Yc      = -4
        n_arr   = 3
        slip    = 1e-5
        Tplot   = [0,t_end/3,t_end] 
    
    
    # PREPARE LEGEND
    if case == 'fig10a':
        col    = ['tab:blue','tab:green','tab:orange']
        orders = (2,0,0)
        n_order = 3
        AI     = (False,False,True)
        fill   = [True,False,False]
        plt.plot([],label='Approach B',color=col[2])
        plt.plot([],label="Lacey's model",color=col[1])
        plt.plot([],label='Reference sol.',color=col[0])
        plt.legend(loc='upper right',fontsize=size_legend)
        style = ['pyblue,thick,fill=pyblue,fill opacity=0.5','pygreen,thick','pyorange,thick']
        mark   = ['ref','Lacey','AI']
    elif case == 'fig10b':
        col    = ['tab:blue','tab:green','tab:orange']
        orders = (2,0,0)
        n_order = 3
        AI     = (False,False,True)
        fill   = [True,False,False]
        style = ['pyblue,thick,fill=pyblue,fill opacity=0.5','pygreen,thick','pyorange,thick']
        mark   = ['ref','Lacey','AI']
    
    
    # Solve
    k = 0
    solr =  np.empty([n_order,n_arr,grid_n])
    solx =  np.empty([n_order,n_arr])
    soly =  np.empty([n_order,n_arr])
    for order in orders:
        drop = ODEdrop(het=g,t_end=t_end,order=order,Xc=Xc,Yc=Yc,slip = slip,n=grid_n,method='RK23',
                       AI=AI[k],model_name = '../model/model_n1600_m16_w128')#'LSODA')
        drop.solve()
        drop.drawcl(Tplot,color=col[k],fill=fill[k])
        solx[k], soly[k], solr[k]= drop.getcl(t=Tplot,coord='polar')
        for t in Tplot:
            out_x,out_y = drop.getcl(t=t,coord='cartesian')
            tikzplot('figures/data_'+case+'_'+mark[k]+'_t='+str(t)+'.txt',np.append(out_x,out_x[0]),np.append(out_y,out_y[0]),style=style[k],forget=False)
        k += 1
   
    x_loc     = solx[2,-1]+ solr[2,-1]*np.cos(u)
    x_loc_ref = solx[0,-1]+ solr[0,-1]*np.cos(u)
    y_loc     = soly[2,-1]+ solr[2,-1]*np.sin(u)
    y_loc_ref = soly[0,-1]+ solr[0,-1]*np.sin(u)
    #Area  = 0.5*np.sum(solr[0,-1]**2)*(u[1]-u[0])
    Area  = calculate_area(x_loc_ref,y_loc_ref) 
    error = frechet(np.array([x_loc,y_loc]).T,np.array([x_loc_ref,y_loc_ref]).T)*np.sqrt(np.pi/Area)
    print(f'auxiliary error = {error:.3%}')
    # 
    # EXPORT FIGURE
    # 
    if case == 'fig10a':
        plt.xlim([-8, +6])
        plt.ylim([-8, +6])
        ax.set_xticks([-6,-4,-2,0,2,4,6])
        ax.set_yticks([-6,-4,-2,0,2,4,6])
        xf,yf = np.meshgrid(np.linspace(-8,+6,1000),np.linspace(-8,+6,1000))
        Gf = g(xf,yf)
        # plt.pcolor(xf[0],yf[:,0],Gf,shading='auto',cmap=cmap)
        ax.imshow(Gf,extent=[-8, 6, -8, 6],cmap=cmap,origin='lower')
    elif case == 'fig10b':    
        plt.xlim([-8, +6])
        plt.ylim([-8, +6])
        ax.set_xticks([-6,-4,-2,0,2,4,6])
        ax.set_yticks([-6,-4,-2,0,2,4,6])
        xf,yf = np.meshgrid(np.linspace(-8,+6,1000),np.linspace(-8,+6,1000))
        Gf = g(xf,yf)
        # plt.pcolor(xf[0],yf[:,0],Gf,shading='auto',cmap=cmap)
        ax.imshow(Gf,extent=[-8, 6, -8, 6],cmap=cmap,origin='lower')
    
    
    matplotlib.image.imsave("figures/het_"+case+".png",Gf,vmin=1,vmax=3,cmap=cmap,origin='lower')
        
    plt.xticks(fontsize = size_ticks)
    plt.yticks(fontsize = size_ticks)
    plt.savefig('figures/{}.png'.format(case),bbox_inches='tight',transparent=True,format='png',dpi=300)
    plt.clf()


    
    

        
    
    
