"""
Figure 9 of JFM 2019 vs PF2021
"""
import sys
sys.path.append("main")
from ODEdrop import *

from matplotlib.colors import LinearSegmentedColormap
import matplotlib
from scipy import interpolate
import scipy.fft as scipy_fft
from similaritymeasures import frechet_dist

def frechet(C1,C2,closed=True):
    if closed:
        n1 = len(C1)
        n2 = len(C2)
        
        if n1>n2:
            n1,n2 = n2,n1
            C1,C2 = C2,C1
            
        C2 = np.vstack((C2,C2[0]))
        i = np.arange(n1)
        q = np.inf
        for j in range(n1):
            ids = np.hstack((np.roll(i,j),[-j]))
            q = np.min([q,frechet_dist(C1[ids],C2)])
            
        return q
    else:
        return frechet_dist(C1,C2)

def calculate_area(x, y):
    n = len(x)
    area = 0

    for i in range(n):
        j = (i + 1) % n
        area += (x[i] * y[j]) - (x[j] * y[i])

    return abs(area) / 2.0


def tikzplot(file,X,Y,style='',forget=True):
   with open(file,"w") as fid:
       if forget:
           style = style+',forget plot'

       if len(style)>0:
           fid.write('\\addplot['+style+']\n')
       else:
           fid.write('\\addplot\n')
       fid.write('table[row sep=crcr]{\n')
       dataXY = [X,Y]
       for x in zip(*dataXY):
           fid.write("%1.8f\t%1.8f\\\\\n" % (x))
       fid.write('};\n')


cases = ['case1','case2','case3','case4','case5','case6']
style = ['pyblue,thick,fill=pyblue,fill opacity=0.5','pyorange,thick']

# PREPARE LEGEND
col    = ['tab:blue','tab:orange']
linestyle = ['-','-']
orders = (2,0)
n_order = 2 
fill=[True,False]
AI = [False,True]
mark   = ['ref','AI']

# Parameters & initialisation
t_end   = 400
grid_n  = 128
u       = np.arange(grid_n)*2*np.pi/grid_n
Xc      = 8
Yc      = 7
Tplot   = [0,t_end] #np.linspace(0.1,1,2) 
n_arr   = 2
    
L=15

def g(d,x,y):
    x = np.cos(d[6])*(x/L) - np.sin(d[6])*(y/L)
    y = np.sin(d[6])*(x/L) + np.cos(d[6])*(y/L)
    return 1 + d[0]*x + d[1]*np.tanh(d[2]*np.cos(d[3]*np.pi*(x*np.sin(d[5])+np.cos(d[5])*y))*\
                                          np.cos(d[4]*np.pi*x))

for case in cases:
    print('===================')
    print(case)
    print('===================')
    if case=='case1':
        het_d = [0.0,0.2,5,20,0,0,0]
    elif case=='case2':
        het_d = [0.8,0.2,5,0,20,0,0]
    elif case=='case3':
        het_d = [0.0,0.2,5,20,20,0,0]
    elif case=='case4':
        het_d = [0.8,0.2,5,20,20,0,0]
    elif case=='case5':
        het_d = [0.8,0.2,5,20,0,np.pi/3,0]
    elif case=='case6':
        het_d = [0.8,0.2,5,20,20,0,np.pi/6]

  
    # DRAW SURFACE
    plt.clf()
    fig, ax = plt.subplots(figsize=(5,5))
    xf,yf = np.meshgrid(np.linspace(0,L,250),np.linspace(0,L,250))
    Gf = g(het_d,xf,yf)
    cmap = LinearSegmentedColormap.from_list('custom greys', [(1,1,1),(0.6,0.6,0.6)], N=256)
    ax.imshow(Gf,extent=[0,L,0,L],cmap=cmap,origin='lower')
    matplotlib.image.imsave("figures/het_"+case+".png",Gf,cmap=cmap,vmin=1,vmax=2,origin='lower')

    
    # Solve
    k = 0
    solr =  np.empty([n_order,n_arr,grid_n])
    solx =  np.empty([n_order,n_arr])
    soly =  np.empty([n_order,n_arr])

    for order in orders:
        drop = ODEdrop(het=g,t_end=t_end,order=order,Xc=Xc,Yc=Yc,het_d=het_d,n=grid_n,method='RK23', 
                       AI = AI[k], model_name = '../model/model_n1600_m16_w128')
        drop.solve()
        drop.drawcl(Tplot,color=col[k],style = linestyle[k],fill=fill[k])
        solx[k], soly[k], solr[k]= drop.getcl(t=Tplot,coord='polar')
        for t in Tplot:
            out_x,out_y = drop.getcl(t=t,coord='cartesian')
            tikzplot('figures/data_'+case+'_'+mark[k]+'_t='+str(t)+'.txt',np.append(out_x,out_x[0]),np.append(out_y,out_y[0]),style=style[k],forget=False)
        k += 1
    
    x_loc     = solx[1,-1]+ solr[1,-1]*np.cos(u)
    x_loc_ref = solx[0,-1]+ solr[0,-1]*np.cos(u)
    y_loc     = soly[1,-1]+ solr[1,-1]*np.sin(u)
    y_loc_ref = soly[0,-1]+ solr[0,-1]*np.sin(u)
    Area  = calculate_area(x_loc_ref,y_loc_ref) 
    error = frechet(np.array([x_loc,y_loc]).T,np.array([x_loc_ref,y_loc_ref]).T)*np.sqrt(np.pi/Area)
    print(f'auxiliary error = {error:.3%}')
    
    # EXPORT FIGURE    
    size_title  = 20
    size_labels = 18
    size_ticks  = 15
    size_legend = 15
    # 
    plt.xlim([0,L])
    plt.ylim([0,L])
    plt.xticks(fontsize = size_ticks)
    plt.yticks(fontsize = size_ticks)
    plt.savefig('figures/{}.png'.format(case),bbox_inches='tight',transparent=True,format='png',dpi=300)
    
